#![allow(clippy::module_name_repetitions)]

use core::{
    convert::TryFrom,
    fmt::{Debug, Formatter, Result as FmtResult},
    ops::Not,
};

use alloc::{boxed::Box, string::String};
use serde::{Deserialize, Serialize};

use crate::utils::{InvalidActionTypeError, NonEmptyStr};

use super::{object_type, GenericParameters, HashMap};

// POST /v1/actions/acknowledge-problem
#[derive(Serialize, Deserialize, Clone, PartialEq, Eq)]
pub struct AcknowledgeProblemAction {
    #[serde(flatten)]
    generics: GenericParameters,
    #[serde(rename = "type")]
    a_type: TargetType,
    author: Box<str>,
    comment: Box<str>,
    #[serde(
        with = "time::serde::timestamp::option",
        skip_serializing_if = "Option::is_none"
    )]
    expiry: Option<time::OffsetDateTime>,
    #[serde(skip_serializing_if = "super::internal_not")]
    sticky: bool,
    #[serde(skip_serializing_if = "super::internal_not")]
    notify: bool,
    #[serde(skip_serializing_if = "super::internal_not")]
    persistent: bool,
}

impl Debug for AcknowledgeProblemAction {
    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
        f.debug_struct("AcknowledgeProblemAction")
            .field("filter", &self.generics.filter)
            .field("filter_vars", &self.generics.filter_vars)
            .field("type", &self.a_type)
            .field("author", &self.author)
            .field("comment", &self.comment)
            .field("expiry", &self.expiry)
            .finish_non_exhaustive()
    }
}

#[must_use]
#[derive(Debug, Clone)]
pub struct AcknowledgeProblemActionBuilder {
    generics: GenericParameters,
    a_type: TargetType,
    author: Box<str>,
    comment: Box<str>,
    expiry: Option<time::OffsetDateTime>,
    sticky: bool,
    notify: bool,
    persistent: bool,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Serialize, Deserialize)]
pub enum TargetType {
    Host,
    Service,
}

impl From<TargetType> for object_type::ObjectType {
    #[inline]
    fn from(value: TargetType) -> Self {
        match value {
            TargetType::Host => Self::Host,
            TargetType::Service => Self::Service,
        }
    }
}

impl TryFrom<object_type::ObjectType> for TargetType {
    type Error = InvalidActionTypeError;

    #[inline]
    fn try_from(value: object_type::ObjectType) -> Result<Self, Self::Error> {
        match value {
            object_type::ObjectType::Host => Ok(Self::Host),
            object_type::ObjectType::Service => Ok(Self::Service),
            value => Err(Self::Error { found: value }),
        }
    }
}

impl AcknowledgeProblemAction {
    /// Create a [`AcknowledgeProblemActionBuilder`] using it's [`new`](AcknowledgeProblemActionBuilder::new) function.
    #[inline]
    pub fn builder(filter: NonEmptyStr, a_type: TargetType) -> AcknowledgeProblemActionBuilder {
        AcknowledgeProblemActionBuilder::new(filter, a_type)
    }

    #[must_use]
    #[inline]
    pub fn filter(&self) -> Option<&str> {
        self.generics.filter.as_deref()
    }

    #[must_use]
    #[inline]
    pub const fn filter_vars(&self) -> Option<&HashMap<String, String>> {
        self.generics.filter_vars.as_ref()
    }

    #[must_use]
    #[inline]
    pub const fn action_type(&self) -> &TargetType {
        &self.a_type
    }

    #[must_use]
    #[inline]
    pub const fn author(&self) -> &str {
        &self.author
    }

    #[must_use]
    #[inline]
    pub const fn comment(&self) -> &str {
        &self.comment
    }

    #[must_use]
    #[inline]
    pub const fn expiry(&self) -> Option<time::OffsetDateTime> {
        self.expiry
    }

    #[must_use]
    #[inline]
    pub const fn is_sticky(&self) -> bool {
        self.sticky
    }

    #[must_use]
    #[inline]
    pub const fn will_notify(&self) -> bool {
        self.notify
    }

    #[must_use]
    #[inline]
    pub const fn is_persistent(&self) -> bool {
        self.persistent
    }
}

impl AcknowledgeProblemActionBuilder {
    /// Create a new builder for [`AcknowledgeProblemAction`].
    #[inline]
    pub fn new(filter: NonEmptyStr, a_type: TargetType) -> AcknowledgeProblemActionBuilder {
        let filter = Some(filter.into());

        AcknowledgeProblemActionBuilder {
            generics: GenericParameters {
                filter,
                ..GenericParameters::default()
            },
            a_type,
            author: Box::<str>::default(),
            comment: Box::<str>::default(),
            expiry: None,
            sticky: false,
            notify: false,
            persistent: false,
        }
    }

    #[inline]
    #[must_use]
    pub fn build(self) -> AcknowledgeProblemAction {
        AcknowledgeProblemAction {
            generics: self.generics,
            a_type: self.a_type,
            author: self.author,
            comment: self.comment,
            expiry: self.expiry,
            notify: self.notify,
            persistent: self.persistent,
            sticky: self.sticky,
        }
    }

    #[inline]
    pub fn filter(mut self, filter: NonEmptyStr) -> Self {
        self.generics.filter = Some(filter.into());
        self
    }

    #[inline]
    pub fn filter_vars(mut self, filter_vars: HashMap<String, String>) -> Self {
        self.generics.filter_vars = filter_vars.is_empty().not().then_some(filter_vars);
        self
    }

    /// Set the action type.
    #[inline]
    pub fn action_type(mut self, a_type: TargetType) -> Self {
        self.a_type = a_type;
        self
    }

    #[inline]
    pub fn author<Author>(mut self, author: Author) -> Self
    where
        Author: Into<Box<str>>,
    {
        self.author = author.into();
        self
    }

    #[inline]
    pub fn comment<Comment>(mut self, comment: Comment) -> Self
    where
        Comment: Into<Box<str>>,
    {
        self.comment = comment.into();
        self
    }

    #[inline]
    pub fn expiry<Expiry>(mut self, expiry: Expiry) -> Self
    where
        Expiry: Into<Option<time::OffsetDateTime>>,
    {
        self.expiry = expiry.into();
        self
    }

    #[inline]
    pub fn sticky(mut self, sticky: bool) -> Self {
        self.sticky = sticky;
        self
    }

    #[inline]
    pub fn notify(mut self, notify: bool) -> Self {
        self.notify = notify;
        self
    }

    #[inline]
    pub fn persistent(mut self, persistent: bool) -> Self {
        self.persistent = persistent;
        self
    }
}
