#![allow(clippy::module_name_repetitions)]

use core::{convert::TryFrom, ops::Not};

use alloc::{boxed::Box, string::String};
use serde::{Deserialize, Serialize};

use crate::utils::{InvalidActionTypeError, NonEmptyStr};

use super::{object_type, GenericParameters, HashMap};

// POST /v1/actions/add-comment
#[derive(Serialize, Deserialize, Debug, PartialEq)]
pub struct AddCommentAction {
    #[serde(flatten)]
    generics: GenericParameters,
    #[serde(rename = "type")]
    a_type: object_type::ObjectType,
    author: Box<str>,
    comment: Box<str>,
    #[serde(
        skip_serializing_if = "Option::is_none",
        with = "time::serde::timestamp::option"
    )]
    expiry: Option<time::OffsetDateTime>,
}

#[derive(Debug, Clone)]
#[must_use]
pub struct AddCommentActionBuilder {
    generics: GenericParameters,
    a_type: object_type::ObjectType,
    author: Box<str>,
    comment: Box<str>,
    expiry: Option<time::OffsetDateTime>,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum TargetType {
    Host,
    Service,
}

impl From<TargetType> for object_type::ObjectType {
    #[inline]
    fn from(value: TargetType) -> Self {
        match value {
            TargetType::Host => Self::Host,
            TargetType::Service => Self::Service,
        }
    }
}

impl TryFrom<object_type::ObjectType> for TargetType {
    type Error = InvalidActionTypeError;

    #[inline]
    fn try_from(value: object_type::ObjectType) -> Result<Self, Self::Error> {
        match value {
            object_type::ObjectType::Host => Ok(Self::Host),
            object_type::ObjectType::Service => Ok(Self::Service),
            value => Err(Self::Error { found: value }),
        }
    }
}

impl AddCommentAction {
    /// Create a [`AddCommentActionBuilder`] using it's [`new`](AddCommentActionBuilder::new) function.
    #[inline]
    pub fn builder(filter: NonEmptyStr, a_type: TargetType) -> AddCommentActionBuilder {
        AddCommentActionBuilder::new(filter, a_type)
    }
}

impl AddCommentActionBuilder {
    /// Create a new builder for [`AddCommentAction`].
    #[inline]
    pub fn new(filter: NonEmptyStr, a_type: TargetType) -> Self {
        let filter = Some(filter.into());
        AddCommentActionBuilder {
            generics: GenericParameters {
                filter,
                ..GenericParameters::default()
            },
            a_type: a_type.into(),
            author: Box::<str>::default(),
            comment: Box::<str>::default(),
            expiry: None,
        }
    }

    #[inline]
    #[must_use]
    pub fn build(self) -> AddCommentAction {
        AddCommentAction {
            a_type: self.a_type,
            author: self.author,
            comment: self.comment,
            expiry: self.expiry,
            generics: self.generics,
        }
    }

    #[inline]
    pub fn filter(mut self, filter: NonEmptyStr) -> Self {
        self.generics.filter = Some(filter.into());
        self
    }

    #[inline]
    pub fn filter_vars(mut self, filter_vars: HashMap<String, String>) -> Self {
        self.generics.filter_vars = filter_vars.is_empty().not().then_some(filter_vars);
        self
    }

    /// Set the action type.
    #[inline]
    pub fn action_type(mut self, a_type: TargetType) -> Self {
        self.a_type = a_type.into();
        self
    }

    #[inline]
    pub fn author<Author>(mut self, author: Author) -> Self
    where
        Author: Into<Box<str>>,
    {
        self.author = author.into();
        self
    }

    #[inline]
    pub fn comment<Comment>(mut self, comment: Comment) -> Self
    where
        Comment: Into<Box<str>>,
    {
        self.comment = comment.into();
        self
    }

    #[inline]
    pub fn expiry<Expiry>(mut self, expiry: Expiry) -> Self
    where
        Expiry: Into<Option<time::OffsetDateTime>>,
    {
        self.expiry = expiry.into();
        self
    }
}
