#[cfg(not(feature = "use-fx-hash"))]
extern crate std;

pub mod acknowledge_problem_action;
pub mod add_comment_action;
pub mod object_type;
pub mod remove_acknowledgement_action;
pub mod remove_comment_action;
pub mod remove_downtime_action;
pub mod reschedule_check_action;
pub mod schedule_downtime_action;
pub mod status;

use alloc::boxed::Box;
use alloc::string::{String, ToString};
use alloc::vec::Vec;
use core::fmt::Formatter;
use core::fmt::{Debug, Result as FmtResult};
#[cfg(feature = "use-fx-hash")]
use core::hash::BuildHasherDefault;
use core::ops::Not;
use core::option::Option;
use serde::{Deserialize, Serialize};

#[cfg(feature = "use-fx-hash")]
type HashMap<K, V> = crate::HashMap<K, V, BuildHasherDefault<rustc_hash::FxHasher>>;
#[cfg(not(feature = "use-fx-hash"))]
type HashMap<K, V> = crate::HashMap<K, V, std::collections::hash_map::RandomState>;

pub use acknowledge_problem_action::AcknowledgeProblemAction;
pub use add_comment_action::AddCommentAction;
pub use remove_acknowledgement_action::{
    RemoveAcknowledgementAction, RemoveAcknowledgementActionBuilder,
};
pub use remove_comment_action::{RemoveCommentAction, RemoveCommentActionBuilder};
pub use remove_downtime_action::{RemoveDowntimeAction, RemoveDowntimeActionBuilder};
pub use reschedule_check_action::{RescheduleCheckAction, RescheduleCheckActionBuilder};
pub use schedule_downtime_action::{ScheduleDowntimeAction, ScheduleDowntimeActionBuilder};

/// Generic parameters for all requests/action types
#[derive(Serialize, Deserialize, Clone, Default, Debug, PartialEq, Eq)]
pub struct GenericParameters {
    pretty: bool,
    //    verbose: bool,
    filter: Option<Box<str>>,
    filter_vars: Option<HashMap<String, String>>,
}

// GET /v1/objects/<type>
/// A request to the object queries
#[derive(Serialize, Deserialize, Default, PartialEq, Eq)]
pub struct QueryRequest {
    #[serde(flatten)]
    generics: GenericParameters,
    attrs: Option<Vec<Box<str>>>,
    joins: Option<Vec<Box<str>>>,
    meta: Option<Vec<Box<str>>>,
}

impl Debug for QueryRequest {
    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
        f.debug_struct("QueryRequest")
            .field("filter", &self.generics.filter)
            .field("filter_vars", &self.generics.filter_vars)
            .field("attrs", &self.attrs)
            .field("joins", &self.joins)
            .field("meta", &self.meta)
            .finish()
    }
}

/// Builder for [`QueryRequest`](self::QueryRequest)
#[derive(Default, Clone)]
#[must_use]
pub struct QueryRequestBuilder {
    generics: GenericParameters,
    attrs: Option<Vec<Box<str>>>,
    joins: Option<Vec<Box<str>>>,
    meta: Option<Vec<Box<str>>>,
}

/// The reply obtained from an object query
#[derive(Serialize, Deserialize, Default, Clone, PartialEq, Eq)]
pub struct QueryReply {
    pub results: Vec<QueryResult>,
}

#[derive(Serialize, Deserialize, Clone, PartialEq, Eq)]
pub struct QueryResult {
    /// The name of the object this [QueryResult](Self) is about.
    pub name: Box<str>,
    /// The type of the object, one of [`ObjectType`](object_type::ObjectType).
    #[serde(rename = "type")]
    pub q_type: object_type::ObjectType,
    /// The attributes returned about the object (Should be a subset of
    /// [`Object`](object_type::Object))
    pub attrs: HashMap<String, String>,
    pub joins: HashMap<String, String>,
}

#[derive(Serialize, Deserialize, Default, Clone, PartialEq)]
pub struct ActionReply {
    pub results: Vec<ActionResult>,
}

#[derive(Serialize, Deserialize, Default, Clone, PartialEq)]
pub struct ActionResult {
    pub code: f64,
    pub status: Box<str>,
}

#[derive(Serialize, Deserialize, Debug, PartialEq)]
pub struct UpdateAction {
    pub attrs: object_type::Object,
}

// POST /v1/actions/process-check-result
#[derive(Serialize, Deserialize, Debug, Default, PartialEq)]
pub struct ProcessCheckResultAction {
    #[serde(flatten)]
    pub generics: GenericParameters,
    #[serde(rename = "type")]
    pub a_type: Box<str>,
    pub exit_status: u64,
    pub plugin_output: Box<str>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub performance_data: Option<Vec<Box<str>>>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub check_command: Option<Vec<Box<str>>>,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub check_source: Option<Box<str>>,
    #[serde(
        skip_serializing_if = "Option::is_none",
        with = "time::serde::timestamp::option"
    )]
    pub execution_start: Option<time::OffsetDateTime>,
    #[serde(
        skip_serializing_if = "Option::is_none",
        with = "time::serde::timestamp::option"
    )]
    pub execution_end: Option<time::OffsetDateTime>,
    pub ttl: f64,
}

// POST /v1/actions/send-custom-notification
#[derive(Serialize, Deserialize, Debug, Default, PartialEq, Eq)]
pub struct SendCustomNotificationAction {
    #[serde(flatten)]
    pub generics: GenericParameters,
    #[serde(rename = "type")]
    pub a_type: Box<str>,
    pub author: Box<str>,
    pub comment: Box<str>,
    pub force: bool,
}

// POST /v1/actions/delay-notification
#[derive(Serialize, Deserialize, Debug, PartialEq, Eq)]
pub struct DelayNotificationAction {
    #[serde(flatten)]
    pub generics: GenericParameters,
    #[serde(rename = "type")]
    pub a_type: object_type::ObjectType,
    #[serde(with = "time::serde::timestamp")]
    pub timestamp: time::OffsetDateTime,
}

// POST /v1/actions/shutdown-process
pub type ShutdownProcessAction = GenericParameters;

// POST /v1/actions/restart-process
pub type RestartProcessAction = GenericParameters;

// GET /v1/status
pub type Status = GenericParameters;

impl QueryRequest {
    pub fn builder() -> QueryRequestBuilder {
        QueryRequestBuilder::default()
    }
}

impl QueryRequestBuilder {
    pub fn new() -> Self {
        QueryRequestBuilder::default()
    }

    #[must_use]
    pub fn build(self) -> QueryRequest {
        QueryRequest {
            generics: self.generics,
            attrs: self.attrs,
            joins: self.joins,
            meta: self.meta,
        }
    }

    pub fn filter<Filter>(mut self, filter: Filter) -> Self
    where
        Filter: Into<Box<str>>,
    {
        let filter: Box<str> = filter.into();
        self.generics.filter = filter.is_empty().not().then_some(filter);
        self
    }

    pub fn filter_vars(mut self, filter_vars: HashMap<String, String>) -> Self {
        self.generics.filter_vars = filter_vars.is_empty().not().then_some(filter_vars);
        self
    }

    pub fn attrs<Attrs>(mut self, attrs: Attrs) -> Self
    where
        Attrs: IntoIterator,
        Attrs::Item: ToString,
    {
        let tmp: Vec<Box<str>> = attrs.into_iter().map(|s| s.to_string().into()).collect();
        self.attrs = tmp.is_empty().not().then_some(tmp);
        self
    }

    pub fn joins<Joins>(mut self, joins: Joins) -> Self
    where
        Joins: IntoIterator,
        Joins::Item: ToString,
    {
        let tmp: Vec<Box<str>> = joins.into_iter().map(|s| s.to_string().into()).collect();
        self.joins = tmp.is_empty().not().then_some(tmp);

        self
    }

    pub fn meta<Meta>(mut self, meta: Meta) -> Self
    where
        Meta: IntoIterator,
        Meta::Item: ToString,
    {
        let tmp: Vec<Box<str>> = meta.into_iter().map(|s| s.to_string().into()).collect();
        self.meta = tmp.is_empty().not().then_some(tmp);
        self
    }
}

// This is due to how the serde(skip_serializing_if = "...") attribute works
#[allow(clippy::trivially_copy_pass_by_ref)]
fn internal_not(b: &bool) -> bool {
    !b
}

#[cfg(test)]
mod test {
    use crate::HashMap;

    use alloc::{format, vec};

    use super::{GenericParameters, QueryRequest, QueryRequestBuilder};

    #[test]
    fn query_request_builder() {
        const FILTER: &str = "regex(^test$, service.name)";
        let expected_filter_vars = {
            let mut tmp = HashMap::default();
            tmp.insert("test".into(), "test".into());
            tmp
        };
        let expected = QueryRequest {
            generics: GenericParameters {
                pretty: false,
                filter: Some(FILTER.into()),
                filter_vars: Some(expected_filter_vars.clone()),
            },
            attrs: Some(vec!["name".into()]),
            joins: None,
            meta: None,
        };

        let builder = QueryRequest::builder()
            .filter(FILTER)
            .attrs(["name"])
            .filter_vars(expected_filter_vars);

        let cloned_build = builder.clone().build();
        let built = builder.build();
        assert_eq!(built, expected);
        assert_eq!(built, cloned_build);
        assert_eq!(
            format!("{expected:?}"),
            format!(
                "QueryRequest {{ filter: {:?}, filter_vars: {:?}, attrs: {:?}, joins: {:?}, meta: {:?} }}",
                &expected.generics.filter,
                &expected.generics.filter_vars,
                &expected.attrs,
                &expected.joins,
                &expected.meta
            )
        );
        assert_eq!(
            QueryRequestBuilder::default().build(),
            QueryRequest::default()
        );
    }
}
