use alloc::{string::String, vec::Vec};

use core::fmt::{Display, Formatter, Result as FmtResult};

use super::HashMap;

#[derive(serde::Serialize, serde::Deserialize, Clone, Copy, Debug, PartialEq, Eq, Hash)]
pub enum ObjectType {
    ApiUser,
    CheckCommand,
    Comment,
    Dependency,
    Downtime,
    Endpoint,
    EventCommand,
    Host,
    HostGroup,
    Notification,
    NotificationCommand,
    ScheduledDowntime,
    Service,
    ServiceGroup,
    TimePeriod,
    User,
    UserGroup,
    Zone,
}

impl Display for ObjectType {
    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
        match *self {
            Self::ApiUser => f.write_str("ApiUser"),
            Self::CheckCommand => f.write_str("CheckCommand"),
            Self::Comment => f.write_str("Comment"),
            Self::Dependency => f.write_str("Dependency"),
            Self::Downtime => f.write_str("Downtime"),
            Self::Endpoint => f.write_str("Endpoint"),
            Self::EventCommand => f.write_str("EventCommand"),
            Self::Host => f.write_str("Host"),
            Self::HostGroup => f.write_str("HostGroup"),
            Self::Notification => f.write_str("Notification"),
            Self::NotificationCommand => f.write_str("NotificationCommand"),
            Self::ScheduledDowntime => f.write_str("ScheduledDowntime"),
            Self::Service => f.write_str("Service"),
            Self::ServiceGroup => f.write_str("ServiceGroup"),
            Self::TimePeriod => f.write_str("TimePeriod"),
            Self::User => f.write_str("User"),
            Self::UserGroup => f.write_str("UserGroup"),
            Self::Zone => f.write_str("Zone"),
        }
    }
}

#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, PartialEq)]
#[serde(untagged)]
pub enum Object {
    ApiUser(ApiUser),
    CheckCommand(CheckCommand),
    Host(Host),
    Service(Service),
    IcingaApplication(IcingaApplication),
}

#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, PartialEq)]
pub struct CommonRuntimeAttributes {
    /// Timestamp when the object was created or modified. Synced throughout cluster nodes.
    version: f64,
    #[serde(rename = "type")]
    o_type: ObjectType,
    /// Original values of object attributes modified at runtime.
    original_attributes: Option<HashMap<String, String>>,
    /// Object is active (e.g. a service being checked).
    active: bool,
    /// Object has been paused at runtime (e.g. IdoMysqlConnection). Defaults to false.
    paused: bool,
    /// Templates imported on object compilation.
    templates: Vec<String>,
    /// Configuration package name this object belongs to. Local configuration is set to _etc,
    /// runtime created objects use _api.
    package: String,
    /// Location information where the configuration files are stored.
    source_location: Option<HashMap<String, String>>,
}

/// `ApiUser` objects are used for authentication against the Icinga 2 API
#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, PartialEq)]
pub struct ApiUser {
    #[serde(flatten)]
    common: CommonRuntimeAttributes,
    /// Password string. Note: This attribute is hidden in API responses.
    #[serde(skip_serializing_if = "Option::is_none")]
    password: Option<String>,
    /// Client Common Name (CN).
    #[serde(skip_serializing_if = "Option::is_none")]
    client_cn: Option<String>,
    /// Array of permissions.
    permissions: Vec<String>,
}

/// A check command definition. Additional default command custom variables can be defined here.
#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, PartialEq)]
pub struct CheckCommand {
    /// The command. This can either be an array of individual command arguments. Alternatively a
    /// string can be specified in which case the shell interpreter (usually /bin/sh) takes care of
    /// parsing the command. When using the “arguments” attribute this must be an array. Can be
    /// specified as function for advanced implementations.
    command: Vec<String>,
    /// A dictionary of macros which should be exported as environment variables prior to executing
    /// the command.
    env: Option<HashMap<String, String>>,
    /// A dictionary containing custom variables that are specific to this command.
    vars: Option<HashMap<String, String>>,
    /// The command timeout in seconds. Defaults to 1m.
    timeout: Option<u64>,
    /// A dictionary of command arguments.
    arguments: Option<HashMap<String, String>>,
}

/// The `IcingaApplication` object is required to start Icinga 2. The object name must be app. If the
/// object configuration is missing, Icinga 2 will automatically create an `IcingaApplication`
/// object.
#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, Default, PartialEq)]
pub struct IcingaApplication {
    /// Whether notifications are globally enabled. Defaults to true.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) enable_notifications: Option<bool>,
    /// Whether event handlers are globally enabled. Defaults to true.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) enable_event_handlers: Option<bool>,
    /// Whether flap detection is globally enabled. Defaults to true.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) enable_flapping: Option<bool>,
    /// Whether active host checks are globally enabled. Defaults to true.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) enable_host_checks: Option<bool>,
    /// Whether active service checks are globally enabled. Defaults to true.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) enable_service_checks: Option<bool>,
    /// Whether performance data processing is globally enabled. Defaults to true.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) enable_perfdata: Option<bool>,
    /// A dictionary containing custom variables that are available globally.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) vars: Option<HashMap<String, String>>,
    /// Specify the Icinga environment. This overrides the Environment constant specified in the
    /// configuration or on the CLI with --define. Defaults to empty.
    #[serde(skip_serializing_if = "Option::is_none")]
    pub(crate) environment: Option<String>,
}

/// `Service` objects describe network services and how they should be checked by Icinga 2.
#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, PartialEq)]
pub struct Service {
    #[serde(flatten)]
    common: CommonRuntimeAttributes,
    display_name: Option<String>,
    host_name: String,
    groups: Option<Vec<String>>,
    vars: Option<HashMap<String, String>>,
    check_command: String,
    max_check_attempts: Option<u64>,
    check_period: Option<String>,
    check_timeout: Option<u64>,
    check_interval: Option<u64>,
    retry_interval: Option<u64>,
    enable_notifications: Option<bool>,
    enable_active_checks: Option<bool>,
    enable_passive_checks: Option<bool>,
    enable_event_handler: Option<bool>,
    enable_flapping: Option<bool>,
    flapping_threshold_high: Option<f64>,
    flapping_threshold_low: Option<f64>,
    flapping_ignore_states: Option<Vec<String>>,
    enable_perfdata: Option<bool>,
    event_command: Option<String>,
    volatile: Option<bool>,
    zone: Option<String>,
    name: String,
    command_endpoint: Option<String>,
    notes: Option<String>,
    notes_url: Option<String>,
    action_url: Option<String>,
    icon_image: Option<String>,
    icon_image_alt: Option<String>,
}

#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, PartialEq)]
pub struct Host {
    #[serde(flatten)]
    common: CommonRuntimeAttributes,
    display_name: Option<String>,
    address: Option<String>,
    address6: Option<String>,
    groups: Option<Vec<String>>,
    vars: Option<HashMap<String, String>>,
    check_command: String,
    max_check_attempts: Option<u64>,
    check_period: Option<String>,
    check_timeout: Option<u64>,
    check_interval: Option<u64>,
    retry_interval: Option<u64>,
    enable_notifications: Option<bool>,
    enable_active_checks: Option<bool>,
    enable_passive_checks: Option<bool>,
    enable_event_handler: Option<bool>,
    enable_flapping: Option<bool>,
    enable_perfdata: Option<bool>,
    event_command: Option<String>,
    flapping_threshold_high: Option<f64>,
    flapping_threshold_low: Option<f64>,
    flapping_ignore_states: Option<Vec<String>>,
    volatile: Option<bool>,
    zone: Option<String>,
    command_endpoint: Option<String>,
    notes: Option<String>,
    notes_url: Option<String>,
    action_url: Option<String>,
    icon_image: Option<String>,
    icon_image_alt: Option<String>,
}
