#![allow(clippy::module_name_repetitions)]

use core::{convert::TryFrom, ops::Not};

use alloc::{boxed::Box, string::String};
use serde::{Deserialize, Serialize};

use crate::utils::{InvalidActionTypeError, NonEmptyStr};

use super::{object_type, GenericParameters, HashMap};

// POST /v1/actions/remove-comment
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq)]
pub struct RemoveCommentAction {
    #[serde(flatten)]
    generics: GenericParameters,
    #[serde(rename = "type")]
    a_type: object_type::ObjectType,
    #[serde(skip_serializing_if = "Option::is_none")]
    author: Option<Box<str>>,
}

#[derive(Debug, Clone)]
#[must_use]
pub struct RemoveCommentActionBuilder {
    generics: GenericParameters,
    a_type: object_type::ObjectType,
    author: Option<Box<str>>,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum TargetType {
    Host,
    Service,
}

impl From<TargetType> for object_type::ObjectType {
    #[inline]
    fn from(value: TargetType) -> Self {
        match value {
            TargetType::Host => Self::Host,
            TargetType::Service => Self::Service,
        }
    }
}

impl TryFrom<object_type::ObjectType> for TargetType {
    type Error = InvalidActionTypeError;

    #[inline]
    fn try_from(value: object_type::ObjectType) -> Result<Self, Self::Error> {
        match value {
            object_type::ObjectType::Host => Ok(Self::Host),
            object_type::ObjectType::Service => Ok(Self::Service),
            value => Err(Self::Error { found: value }),
        }
    }
}

impl RemoveCommentAction {
    /// Create a [`RemoveCommentActionBuilder`] using it's [`new`](RemoveCommentActionBuilder::new) function.
    #[inline]
    pub fn builder(filter: NonEmptyStr, a_type: TargetType) -> RemoveCommentActionBuilder {
        RemoveCommentActionBuilder::new(filter, a_type)
    }
}

impl RemoveCommentActionBuilder {
    /// Create a new builder for [`RemoveCommentAction`].
    #[inline]
    pub fn new(filter: NonEmptyStr, action_type: TargetType) -> Self {
        let filter = Some(filter.into());
        RemoveCommentActionBuilder {
            generics: GenericParameters {
                filter,
                ..GenericParameters::default()
            },
            a_type: action_type.into(),
            author: None,
        }
    }

    #[inline]
    #[must_use]
    pub fn build(self) -> RemoveCommentAction {
        RemoveCommentAction {
            a_type: self.a_type,
            author: self.author,
            generics: self.generics,
        }
    }

    #[inline]
    pub fn filter(mut self, filter: NonEmptyStr) -> Self {
        self.generics.filter = Some(filter.into());
        self
    }

    #[inline]
    pub fn filter_vars(mut self, filter_vars: HashMap<String, String>) -> Self {
        self.generics.filter_vars = filter_vars.is_empty().not().then_some(filter_vars);
        self
    }

    /// Set the action type.
    #[inline]
    pub fn action_type(mut self, action_type: TargetType) -> Self {
        self.a_type = action_type.into();
        self
    }

    #[inline]
    pub fn author<Author>(mut self, author: Author) -> Self
    where
        Author: Into<Box<str>>,
    {
        let author = author.into();
        self.author = author.is_empty().not().then_some(author);
        self
    }
}
