#![allow(clippy::module_name_repetitions)]

use core::{convert::TryFrom, ops::Not};

use alloc::string::String;
use serde::{Deserialize, Serialize};

use crate::utils::{InvalidActionTypeError, NonEmptyStr};

use super::{object_type, GenericParameters, HashMap};

// POST /v1/actions/reschedule-check
#[derive(Serialize, Deserialize, Debug, PartialEq, Eq)]
pub struct RescheduleCheckAction {
    #[serde(flatten)]
    generics: GenericParameters,
    #[serde(rename = "type")]
    a_type: object_type::ObjectType,
    #[serde(
        skip_serializing_if = "Option::is_none",
        with = "time::serde::timestamp::option"
    )]
    next_check: Option<time::OffsetDateTime>,
    force: bool,
}

#[derive(Debug, Clone)]
#[must_use]
pub struct RescheduleCheckActionBuilder {
    generics: GenericParameters,
    a_type: object_type::ObjectType,
    next_check: Option<time::OffsetDateTime>,
    force: bool,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
/// Valid [`ObjectType`](crate::api::object_type::ObjectType) to use as target for [`RescheduleCheckAction`]
pub enum TargetType {
    Host,
    Service,
}

impl From<TargetType> for object_type::ObjectType {
    fn from(value: TargetType) -> Self {
        match value {
            TargetType::Host => Self::Host,
            TargetType::Service => Self::Service,
        }
    }
}

impl TryFrom<object_type::ObjectType> for TargetType {
    type Error = InvalidActionTypeError;

    fn try_from(value: object_type::ObjectType) -> Result<Self, Self::Error> {
        match value {
            object_type::ObjectType::Host => Ok(Self::Host),
            object_type::ObjectType::Service => Ok(Self::Service),
            value => Err(Self::Error { found: value }),
        }
    }
}

impl RescheduleCheckAction {
    /// Create a [`RescheduleCheckActionBuilder`] using it's [`new`](RescheduleCheckActionBuilder::new) function.
    ///
    /// # Errors
    ///
    /// This function will return an error if [`RescheduleCheckActionBuilder::new`] return an error.
    pub fn builder(target_type: TargetType, filter: NonEmptyStr) -> RescheduleCheckActionBuilder {
        self::RescheduleCheckActionBuilder::new(target_type, filter)
    }
}

impl RescheduleCheckActionBuilder {
    /// Create a new builder for [`RescheduleCheckAction`].
    ///
    /// # Errors
    ///
    /// This function will return an error if `action_type` is anything but [`Host`](object_type::ObjectType::Host) or [`Service`](object_type::ObjectType::Service).
    pub fn new(target_type: TargetType, filter: NonEmptyStr) -> Self {
        let filter = Some(filter.into());
        Self {
            generics: GenericParameters {
                filter,
                ..GenericParameters::default()
            },
            a_type: target_type.into(),
            next_check: None,
            force: false,
        }
    }

    #[must_use]
    pub fn build(self) -> RescheduleCheckAction {
        RescheduleCheckAction {
            generics: self.generics,
            a_type: self.a_type,
            next_check: self.next_check,
            force: self.force,
        }
    }

    pub fn filter(mut self, filter: NonEmptyStr) -> Self {
        self.generics.filter = Some(filter.into());
        self
    }

    pub fn filter_vars(mut self, filter_vars: HashMap<String, String>) -> Self {
        self.generics.filter_vars = filter_vars.is_empty().not().then_some(filter_vars);
        self
    }

    /// Set the action type.
    ///
    /// # Errors
    ///
    /// This function will return an error if `action_type` is anything but [`Host`](object_type::ObjectType::Host) or [`Service`](object_type::ObjectType::Service).
    pub fn action_type(mut self, a_type: TargetType) -> Self {
        self.a_type = a_type.into();
        self
    }

    #[cfg(not(test))]
    pub fn next_check(mut self, time: time::OffsetDateTime) -> Self {
        self.next_check = Some(time).filter(|time| time.ge(&time::OffsetDateTime::now_utc()));
        self
    }

    #[cfg(test)]
    pub fn next_check(mut self, time: time::OffsetDateTime) -> Self {
        self.next_check = Some(time);
        self
    }

    pub const fn force(mut self, force: bool) -> Self {
        self.force = force;
        self
    }
}

#[cfg(test)]
mod test {
    use alloc::string::String;
    use core::convert::TryFrom;

    use crate::HashMap;

    use crate::{
        api::{reschedule_check_action::TargetType, GenericParameters, RescheduleCheckAction},
        utils::NonEmptyStr,
    };

    #[test]
    fn builder() {
        const FILTER: &str = "filter";
        const ACTION_TYPE: TargetType = TargetType::Service;
        const NEXT_CHECK: time::OffsetDateTime = time::OffsetDateTime::UNIX_EPOCH;
        let filter_vars = {
            let mut tmp = HashMap::default();
            tmp.insert(String::from("test"), String::from("test"));
            tmp
        };
        let filter = NonEmptyStr::try_from(FILTER).unwrap();

        let expected = RescheduleCheckAction {
            generics: GenericParameters {
                pretty: false,
                filter: Some(FILTER.into()),
                filter_vars: filter_vars.clone().into(),
            },
            a_type: ACTION_TYPE.into(),
            next_check: Some(NEXT_CHECK),
            force: true,
        };

        let builder = RescheduleCheckAction::builder(ACTION_TYPE, filter)
            .filter_vars(filter_vars)
            .next_check(NEXT_CHECK)
            .force(true);
        let cloned_build = builder.clone().build();
        let built = builder.build();

        assert_eq!(cloned_build, built);
        assert_eq!(built, expected);
    }
}
