#![allow(clippy::module_name_repetitions)]

use core::{convert::TryFrom, ops::Not};

use alloc::{boxed::Box, string::String};
use serde::{Deserialize, Serialize};

use crate::utils::{InvalidActionTypeError, NonEmptyStr, RequestBuilderError};

use super::{object_type, GenericParameters, HashMap};

// POST /v1/actions/schedule-downtime
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq)]
pub struct ScheduleDowntimeAction {
    #[serde(flatten)]
    generics: GenericParameters,
    #[serde(rename = "type")]
    a_type: object_type::ObjectType,
    author: Box<str>,
    comment: Box<str>,
    #[serde(with = "time::serde::timestamp")]
    start_time: time::OffsetDateTime,
    #[serde(with = "time::serde::timestamp")]
    end_time: time::OffsetDateTime,
    fixed: bool,
    // required if fixed == false
    #[serde(skip_serializing_if = "Option::is_none")]
    duration: Option<f64>,
    #[serde(skip_serializing_if = "Option::is_none")]
    all_services: Option<bool>,
    #[serde(skip_serializing_if = "Option::is_none")]
    trigger_name: Option<Box<str>>,
    #[serde(skip_serializing_if = "Option::is_none")]
    child_option: Option<Box<str>>,
}

#[derive(Debug, Clone)]
#[must_use]
pub struct ScheduleDowntimeActionBuilder {
    generics: GenericParameters,
    a_type: object_type::ObjectType,
    author: Box<str>,
    comment: Box<str>,
    start_time: time::OffsetDateTime,
    end_time: time::OffsetDateTime,
    fixed: bool,
    duration: Option<f64>,
    all_services: Option<bool>,
    trigger_name: Option<Box<str>>,
    child_option: Option<Box<str>>,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum TargetType {
    Host,
    Service,
}

impl From<TargetType> for object_type::ObjectType {
    #[inline]
    fn from(value: TargetType) -> Self {
        match value {
            TargetType::Host => Self::Host,
            TargetType::Service => Self::Service,
        }
    }
}

impl TryFrom<object_type::ObjectType> for TargetType {
    type Error = InvalidActionTypeError;

    #[inline]
    fn try_from(value: object_type::ObjectType) -> Result<Self, Self::Error> {
        match value {
            object_type::ObjectType::Host => Ok(Self::Host),
            object_type::ObjectType::Service => Ok(Self::Service),
            value => Err(Self::Error { found: value }),
        }
    }
}

impl ScheduleDowntimeAction {
    /// Create a [`ScheduleDowntimeActionBuilder`] using it's [`new`](ScheduleDowntimeActionBuilder::new) function.
    ///
    /// # Errors
    ///
    /// This function will return an error if [`ScheduleDowntimeActionBuilder::new`] return an error.
    #[inline]
    pub fn builder(
        author: NonEmptyStr,
        comment: NonEmptyStr,
        end_time: time::OffsetDateTime,
        filter: NonEmptyStr,
        a_type: TargetType,
    ) -> Result<ScheduleDowntimeActionBuilder, RequestBuilderError> {
        ScheduleDowntimeActionBuilder::new(author, comment, end_time, filter, a_type)
    }
}

impl ScheduleDowntimeActionBuilder {
    /// Create a builder for [`ScheduleDowntimeAction`].
    ///
    /// # Errors
    ///
    /// This function will return an error if:
    ///
    /// * `action_type` is anything but [`Host`](object_type::ObjectType::Host) or [`Service`](object_type::ObjectType::Service),
    /// * `end_time` is in the past
    pub fn new(
        author: NonEmptyStr,
        comment: NonEmptyStr,
        end_time: time::OffsetDateTime,
        filter: NonEmptyStr,
        action_type: TargetType,
    ) -> Result<Self, RequestBuilderError> {
        if end_time < time::OffsetDateTime::now_utc() {
            return Err(RequestBuilderError::EndBeforeStart {
                end: end_time,
                start: time::OffsetDateTime::now_utc(),
            });
        }

        let filter = Some(filter.into());
        let author = author.into();
        let comment = comment.into();
        let end_time = end_time;
        let builder = ScheduleDowntimeActionBuilder {
            generics: GenericParameters {
                filter,
                ..GenericParameters::default()
            },
            a_type: action_type.into(),
            author,
            comment,
            start_time: time::OffsetDateTime::now_utc(),
            end_time,
            fixed: true,
            duration: None,
            all_services: None,
            trigger_name: None,
            child_option: None,
        };
        Ok(builder)
    }

    #[inline]
    #[must_use]
    pub fn build(self) -> ScheduleDowntimeAction {
        ScheduleDowntimeAction {
            a_type: self.a_type,
            all_services: self.all_services,
            author: self.author,
            child_option: self.child_option,
            comment: self.comment,
            duration: self.duration,
            end_time: self.end_time,
            fixed: self.fixed,
            generics: self.generics,
            start_time: self.start_time,
            trigger_name: self.trigger_name,
        }
    }

    #[inline]
    pub fn filter(mut self, filter: NonEmptyStr) -> Self {
        self.generics.filter = Some(filter.into());
        self
    }

    #[inline]
    pub fn filter_vars(mut self, filter_vars: HashMap<String, String>) -> Self {
        self.generics.filter_vars = filter_vars.is_empty().not().then_some(filter_vars);
        self
    }

    /// Set the action type.
    #[inline]
    pub fn action_type(mut self, a_type: TargetType) -> Self {
        self.a_type = a_type.into();
        self
    }

    #[inline]
    pub fn author(mut self, author: NonEmptyStr) -> Self {
        self.author = author.into();
        self
    }

    #[inline]
    pub fn comment(mut self, comment: NonEmptyStr) -> Self {
        self.comment = comment.into();
        self
    }

    /// Set when the scheduled downtime will start.
    ///
    /// # Errors
    ///
    /// This function will return an error if `start_time` is after the end of the downtime that would be scheduled.
    pub fn start_time(
        mut self,
        start_time: time::OffsetDateTime,
    ) -> Result<Self, RequestBuilderError> {
        if start_time > self.end_time {
            return Err(RequestBuilderError::StartAfterEnd {
                start: start_time,
                end: self.end_time,
            });
        };
        self.start_time = start_time;
        Ok(self)
    }

    /// Set when the scheduled downtime will end.
    ///
    /// # Errors
    ///
    /// This function will return an error if `end_time` is before the start of the downtime that would be scheduled.
    pub fn end_time(mut self, end_time: time::OffsetDateTime) -> Result<Self, RequestBuilderError> {
        if end_time < self.start_time {
            return Err(RequestBuilderError::EndBeforeStart {
                end: end_time,
                start: self.start_time,
            });
        }
        self.end_time = end_time;
        Ok(self)
    }

    #[inline]
    pub fn flexible(mut self, duration: time::Duration) -> Self {
        self.fixed = false;
        self.duration = Some(duration.as_seconds_f64());
        self
    }

    #[inline]
    pub fn fixed(mut self) -> Self {
        self.fixed = true;
        self.duration = None;
        self
    }

    #[inline]
    pub fn all_services(mut self, b: bool) -> Self {
        self.all_services = Some(b);
        self
    }

    #[inline]
    pub fn trigger_name<TriggerName>(mut self, trigger_name: TriggerName) -> Self
    where
        TriggerName: Into<Box<str>>,
    {
        self.trigger_name = Some(trigger_name.into());
        self
    }

    #[inline]
    pub fn child_option<ChildOption>(mut self, child_option: ChildOption) -> Self
    where
        ChildOption: Into<Box<str>>,
    {
        self.child_option = Some(child_option.into());
        self
    }
}
