use core::fmt::{Debug, Formatter, Result as FmtResult};

use crate::api::status::icinga_application::Response as IcingaApplicationStatus;
use crate::api::{ActionReply, QueryReply};

use alloc::string::String;
use alloc::{boxed::Box, format};
use http_auth_basic::Credentials;
use hyper::client::connect::Connect;
use hyper::{Body, Client, Method, Request};

pub mod error;

#[cfg(not(any(feature = "http1", feature = "http2")))]
compile_error!("Need `http1` or `http2` feature.");

struct Auth {
    inner: Credentials,
}

impl Debug for Auth {
    fn fmt(&self, fmt: &mut Formatter) -> FmtResult {
        fmt.debug_struct("IcingaAuth")
            .field("username", &self.inner.user_id)
            .field("password", &format_args!("Sensitive"))
            .finish()
    }
}

impl Default for Auth {
    fn default() -> Self {
        Self {
            inner: Credentials {
                user_id: String::default(),
                password: String::default(),
            },
        }
    }
}

impl Clone for Auth {
    fn clone(&self) -> Self {
        Self {
            inner: Credentials {
                user_id: self.inner.user_id.clone(),
                password: self.inner.password.clone(),
            },
        }
    }
}

impl Auth {
    fn as_http_header(&self) -> String {
        self.inner.as_http_header()
    }
}

#[derive(Clone, Debug)]
pub struct Service<Connector> {
    url: Box<str>,
    auth: Auth,
    name: Box<str>,
    client: Client<Connector>,
}

#[must_use]
#[derive(Clone, Debug)]
pub struct ServiceBuilder<Connector> {
    url: Box<str>,
    auth: Auth,
    name: Box<str>,
    client: Client<Connector>,
}

impl<Connector> Service<Connector>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    #[must_use]
    pub fn name(&self) -> &str {
        &self.name
    }

    #[must_use]
    pub fn url(&self) -> &str {
        &self.url
    }

    fn build_query(
        &self,
        url: &str,
        query: &crate::api::QueryRequest,
    ) -> Result<Request<hyper::Body>, error::RequestBuildError> {
        Request::builder()
            .method(Method::POST)
            .uri(url)
            .header(hyper::header::ACCEPT, "application/json")
            .header(hyper::header::CONTENT_TYPE, "application/json")
            .header(hyper::header::AUTHORIZATION, self.auth.as_http_header())
            .header("X-HTTP-Method-Override", "GET")
            .body(serde_json::to_vec(query)?.into())
            .map_err(Into::into)
    }

    #[tracing::instrument(level = "debug")]
    pub async fn query_service(
        &self,
        query: &crate::api::QueryRequest,
    ) -> Result<QueryReply, error::Error> {
        let url = format!("{}/v1/objects/services", self.url);

        let response = self.client.request(self.build_query(&url, query)?).await?;
        if !response.status().is_success() {
            return Err(error::Error::StatusCode(response.status()));
        }
        let response = {
            let mut response = response;
            hyper::body::to_bytes(response.body_mut()).await?
        };
        serde_json::from_slice(&response).map_err(Into::into)
    }

    #[tracing::instrument(level = "debug")]
    pub async fn query_hosts(
        &self,
        query: &crate::api::QueryRequest,
    ) -> Result<QueryReply, error::Error> {
        let url = format!("{}/v1/objects/hosts", self.url);

        let response = self.client.request(self.build_query(&url, query)?).await?;
        if !response.status().is_success() {
            return Err(error::Error::StatusCode(response.status()));
        }
        let response = {
            let mut response = response;
            hyper::body::to_bytes(response.body_mut()).await?
        };
        let json: QueryReply = serde_json::from_slice(&response)?;
        Ok(json)
    }

    fn build_action<T: serde::Serialize>(
        &self,
        url: &str,
        action: &T,
    ) -> Result<Request<hyper::Body>, error::RequestBuildError> {
        Request::builder()
            .method(Method::POST)
            .uri(url)
            .header(hyper::header::ACCEPT, "application/json")
            .header(hyper::header::CONTENT_TYPE, "application/json")
            .header(hyper::header::AUTHORIZATION, self.auth.as_http_header())
            .body(serde_json::to_vec(action)?.into())
            .map_err(error::RequestBuildError::from)
    }

    async fn build_and_execute_action<Action: serde::Serialize>(
        &self,
        url: &str,
        action: &Action,
    ) -> Result<ActionReply, error::Error> {
        let response = self.client.request(self.build_action(url, action)?).await?;
        let status = response.status();
        let response = {
            let mut response = response;
            hyper::body::to_bytes(response.body_mut()).await?
        };
        let json: Result<ActionReply, _> = serde_json::from_slice(&response);
        if !status.is_success() {
            if let Ok(json) = json {
                if let Some(result) = {
                    let mut results = json.results;
                    results.pop()
                } {
                    return Err(error::Error::ActionStatusCode {
                        http_status: status,
                        api_status: result.code,
                        message: result.status,
                    });
                }
            }
            return Err(error::Error::StatusCode(status));
        }
        json.map_err(Into::into)
    }

    #[tracing::instrument(level = "debug")]
    pub async fn reschedule_check(
        &self,
        action: &crate::api::RescheduleCheckAction,
    ) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/actions/reschedule-check", self.url);

        self.build_and_execute_action(&url, action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn acknowledge_problem(
        &self,
        action: &crate::api::AcknowledgeProblemAction,
    ) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/actions/acknowledge-problem", self.url);

        self.build_and_execute_action(&url, action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn remove_acknowledgement(
        &self,
        action: &crate::api::RemoveAcknowledgementAction,
    ) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/actions/remove-acknowledgement", self.url);

        self.build_and_execute_action(&url, action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn add_comment(
        &self,
        action: &crate::api::AddCommentAction,
    ) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/actions/add-comment", self.url);

        self.build_and_execute_action(&url, action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn remove_comment(
        &self,
        action: &crate::api::RemoveCommentAction,
    ) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/actions/remove-comment", self.url);

        self.build_and_execute_action(&url, action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn schedule_downtime(
        &self,
        action: &crate::api::ScheduleDowntimeAction,
    ) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/actions/schedule-downtime", self.url);

        self.build_and_execute_action(&url, action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn remove_downtime(
        &self,
        action: &crate::api::RemoveDowntimeAction,
    ) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/actions/remove-downtime", self.url);

        self.build_and_execute_action(&url, action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn set_notification(&self, active: bool) -> Result<ActionReply, error::Error> {
        let url = format!("{}/v1/objects/icingaapplications/app", self.url);

        let app = crate::api::object_type::IcingaApplication {
            enable_notifications: Some(active),
            ..crate::api::object_type::IcingaApplication::default()
        };

        let action = crate::api::UpdateAction {
            attrs: crate::api::object_type::Object::IcingaApplication(app),
        };

        self.build_and_execute_action(&url, &action).await
    }

    #[tracing::instrument(level = "debug")]
    pub async fn get_application_status(
        &self,
    ) -> Result<IcingaApplicationStatus, error::ApplicationStatusError> {
        let url = format!("{}/v1/status/IcingaApplication", self.url);

        let request = Request::builder()
            .method(Method::GET)
            .uri(url)
            .header(hyper::header::ACCEPT, "application/json")
            .header(hyper::header::AUTHORIZATION, self.auth.as_http_header())
            .body(Body::empty())?;
        let response = self.client.request(request).await?;
        if !response.status().is_success() {
            return Err(error::ApplicationStatusError::StatusCode(response.status()));
        }
        let response = {
            let mut response = response;
            hyper::body::to_bytes(response.body_mut()).await?
        };
        serde_json::from_slice(&response).map_err(Into::into)
    }
}

impl<Connector> ServiceBuilder<Connector>
where
    Connector: Connect + Clone,
{
    pub fn new(client: Client<Connector>) -> Self {
        Self {
            client,
            url: Box::default(),
            auth: Auth::default(),
            name: Box::default(),
        }
    }

    pub fn url<Url>(&mut self, url: Url) -> &mut Self
    where
        Url: Into<Box<str>>,
    {
        self.url = url.into();
        self
    }

    pub fn auth<Username, Password>(&mut self, username: Username, password: Password) -> &mut Self
    where
        Username: Into<String>,
        Password: Into<String>,
    {
        let inner = Credentials {
            user_id: username.into(),
            password: password.into(),
        };
        self.auth = Auth { inner };
        self
    }

    pub fn name<Name>(&mut self, service_name: Name) -> &mut Self
    where
        Name: Into<Box<str>>,
    {
        self.name = service_name.into();
        self
    }

    #[must_use]
    pub fn build(&self) -> Option<Service<Connector>> {
        if self.url.is_empty()
            || self.auth.inner.user_id.is_empty()
            || self.auth.inner.password.is_empty()
        {
            return None;
        }

        Some(Service {
            url: self.url.clone(),
            auth: self.auth.clone(),
            name: self.name.clone(),
            client: self.client.clone(),
        })
    }
}
