extern crate std;

use alloc::{boxed::Box, string::String};
use core::fmt::{Formatter, Result as StdResult};
use core::{convert::TryFrom, fmt::Display};
use thiserror::Error as ThisError;

#[derive(Debug, Clone)]
pub struct NonEmptyStr {
    string: Box<str>,
}

impl TryFrom<Box<str>> for NonEmptyStr {
    type Error = Error;

    fn try_from(s: Box<str>) -> Result<Self, Self::Error> {
        if s.is_empty() {
            return Err(Self::Error::EmptyStrError);
        }
        Ok(Self { string: s })
    }
}

impl TryFrom<String> for NonEmptyStr {
    type Error = Error;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        if value.is_empty() {
            return Err(Self::Error::EmptyStrError);
        }
        Ok(Self {
            string: value.into(),
        })
    }
}

impl TryFrom<&str> for NonEmptyStr {
    type Error = Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        if value.is_empty() {
            return Err(Self::Error::EmptyStrError);
        }
        Ok(Self {
            string: value.into(),
        })
    }
}

impl Display for NonEmptyStr {
    fn fmt(&self, f: &mut Formatter<'_>) -> StdResult {
        self.string.fmt(f)
    }
}

impl AsRef<str> for NonEmptyStr {
    #[inline]
    fn as_ref(&self) -> &str {
        self.string.as_ref()
    }
}

impl From<NonEmptyStr> for Box<str> {
    #[inline]
    fn from(s: NonEmptyStr) -> Self {
        s.string
    }
}

// impl From<&NonEmptyStr> for Box<str> {
//     fn from(s: &NonEmptyStr) -> Self {
//         s.string.clone()
//     }
// }

#[derive(ThisError, Debug, Clone, Copy)]
pub enum Error {
    #[error("String is empty")]
    EmptyStrError,
    #[error(transparent)]
    RequestBuilderError(#[from] RequestBuilderError),
    #[error(transparent)]
    InvalidActionTypeError(#[from] InvalidActionTypeError),
}

#[derive(ThisError, Debug, Clone, Copy)]
pub enum RequestBuilderError {
    #[error(transparent)]
    InvalidActionType(#[from] InvalidActionTypeError),
    #[error("Start time {start} after end time {end}")]
    StartAfterEnd {
        end: time::OffsetDateTime,
        start: time::OffsetDateTime,
    },
    #[error("End time {end} before start time {start}")]
    EndBeforeStart {
        end: time::OffsetDateTime,
        start: time::OffsetDateTime,
    },
}

#[derive(ThisError, Debug, Clone, Copy)]
pub struct InvalidActionTypeError {
    pub(crate) found: super::api::object_type::ObjectType,
}

impl Display for InvalidActionTypeError {
    fn fmt(&self, f: &mut Formatter<'_>) -> StdResult {
        write!(f, "Invalid action type (found {})", self.found)
    }
}
